"""
Tests for Name Affirmation signals
"""

import ddt

from django.contrib.auth import get_user_model
from django.test import TestCase

from edx_name_affirmation.models import VerifiedName
from edx_name_affirmation.signals import _listen_for_lms_retire_verified_names

User = get_user_model()


@ddt.ddt
class RetirementSignalVerifiedNamesTest(TestCase):
    """
    Tests for the LMS User Retirement signal for Verified Names
    """

    def setUp(self):
        self.user = User(username='tester', email='tester@test.com')
        self.user.save()
        self.name = 'Jonathan Smith'
        self.profile_name = 'Jon Smith'
        self.idv_attempt_id = 1111111
        self.verified_name_obj = VerifiedName.objects.create(
            user=self.user,
            verified_name=self.name,
            profile_name=self.profile_name,
            verification_attempt_id=self.idv_attempt_id
        )

        self.other_user = User(username='other_tester', email='other_tester@test.com')
        self.other_user.save()
        self.other_name = 'Jonathan Other'
        self.other_profile_name = 'Jon Other'
        self.other_idv_attempt_id = 1111112
        self.verified_name_obj = VerifiedName.objects.create(
            user=self.other_user,
            verified_name=self.other_name,
            profile_name=self.other_profile_name,
            verification_attempt_id=self.other_idv_attempt_id
        )

    def test_retirement_signal(self):
        _listen_for_lms_retire_verified_names(sender=self.__class__, user=self.user)
        self.assertEqual(len(VerifiedName.objects.filter(user=self.user)), 0)
        self.assertEqual(len(VerifiedName.objects.filter(user=self.other_user)), 1)

    def test_retirement_signal_no_verified_names(self):
        no_verified_user = User(username='no_verified', email='no_verified@test.com')
        _listen_for_lms_retire_verified_names(sender=self.__class__, user=no_verified_user)
        self.assertEqual(len(VerifiedName.objects.all()), 2)

    def test_retirement_signal_all_verified_names_for_user(self):
        # create a second verified name for user to check that both names are deleted
        VerifiedName.objects.create(
            user=self.user,
            verified_name='J Smith',
            profile_name=self.profile_name,
            verification_attempt_id=1111112
        )
        _listen_for_lms_retire_verified_names(sender=self.__class__, user=self.user)
        self.assertEqual(len(VerifiedName.objects.filter(user=self.user)), 0)



# Generated by Django 4.2.16 on 2024-09-19 12:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edx_name_affirmation', '0008_alter_historicalverifiedname_options'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='historicalverifiedname',
            options={'get_latest_by': 'history_date', 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical verified name'},
        ),
        migrations.AddField(
            model_name='historicalverifiedname',
            name='platform_verification_attempt_id',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='verifiedname',
            name='platform_verification_attempt_id',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
    ]



"""
Name Affirmation signals
"""

from django.dispatch import Signal
from django.dispatch.dispatcher import receiver

try:
    from openedx.core.djangoapps.user_api.accounts.signals import USER_RETIRE_LMS_MISC
except ImportError:
    # An ImportError should only be raised in tests, where the code is not running as an installation of
    # edx-platform. In this case, the import should default to a generic Signal.
    USER_RETIRE_LMS_MISC = Signal()

from .models import VerifiedName

VERIFIED_NAME_APPROVED = Signal()


@receiver(USER_RETIRE_LMS_MISC)
def _listen_for_lms_retire_verified_names(sender, **kwargs):  # pylint: disable=unused-argument
    user = kwargs.get('user')
    VerifiedName.retire_user(user.id)



"""
Name Affirmation signal handlers
"""

import logging

from openedx_events.learning.signals import (
    IDV_ATTEMPT_APPROVED,
    IDV_ATTEMPT_CREATED,
    IDV_ATTEMPT_DENIED,
    IDV_ATTEMPT_PENDING
)

from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.dispatch.dispatcher import receiver

from edx_name_affirmation.models import VerifiedName
from edx_name_affirmation.signals import VERIFIED_NAME_APPROVED
from edx_name_affirmation.statuses import VerifiedNameStatus
from edx_name_affirmation.tasks import (
    delete_verified_name_task,
    idv_update_verified_name_task,
    proctoring_update_verified_name_task
)

User = get_user_model()

log = logging.getLogger(__name__)


@receiver(post_save, sender=VerifiedName)
def verified_name_approved(sender, instance, **kwargs):  # pylint: disable=unused-argument
    """
    Emit a signal when a verified name's status is updated to "approved".
    """
    if instance.status == VerifiedNameStatus.APPROVED:
        VERIFIED_NAME_APPROVED.send(
          sender='name_affirmation',
          user_id=instance.user.id,
          profile_name=instance.profile_name
        )


@receiver(IDV_ATTEMPT_APPROVED)
@receiver(IDV_ATTEMPT_CREATED)
@receiver(IDV_ATTEMPT_DENIED)
@receiver(IDV_ATTEMPT_PENDING)
def handle_idv_event(sender, signal, **kwargs):  # pylint: disable=unused-argument
    """
    Trigger update to verified names based on open edX IDV events.
    """
    event_data = kwargs.get('idv_attempt')
    user = User.objects.get(id=event_data.user.id)

    # If the user has a pending name change, use that as the full name
    try:
        user_full_name = user.pending_name_change
    except AttributeError:
        user_full_name = event_data.user.pii.name

    status = None
    if signal == IDV_ATTEMPT_APPROVED:
        status = VerifiedNameStatus.APPROVED
    elif signal == IDV_ATTEMPT_CREATED:
        status = VerifiedNameStatus.PENDING
    elif signal == IDV_ATTEMPT_PENDING:
        status = VerifiedNameStatus.SUBMITTED
    elif signal == IDV_ATTEMPT_DENIED:
        status = VerifiedNameStatus.DENIED
    else:
        log.info(f'IDV_ATTEMPT {signal} signal not recognized')  # driven by receiver decorator so should never happen
        return

    log.info(f'IDV_ATTEMPT {status} signal triggering Celery task for user {user.id} '
             f'with name {event_data.name}')
    idv_update_verified_name_task.delay(
        event_data.attempt_id,
        user.id,
        status,
        event_data.name,
        user_full_name,
    )


def platform_verification_delete_handler(sender, instance, signal, **kwargs):  # pylint: disable=unused-argument
    """
    Receiver for VerificationAttempt deletions
    """
    platform_verification_attempt_id = instance.id
    log.info(
        'VerifiedName: platform_verification_delete_handler triggering Celery task for '
        'platform_verification_attempt_id=%(platform_verification_attempt_id)s',
        {
            'platform_verification_attempt_id': platform_verification_attempt_id,
        }
    )
    delete_verified_name_task.delay(platform_verification_attempt_id, None)


def proctoring_attempt_handler(
    attempt_id,
    user_id,
    status,
    full_name,
    profile_name,
    is_practice_exam,
    is_proctored,
    backend_supports_onboarding,
    **kwargs
):
    """
    Receiver for proctored exam attempt updates.

    Args:
        attempt_id(int): ID associated with the proctored exam attempt
        user_id(int): ID associated with the proctored exam attempt's user
        status(str): status in proctoring language for the proctored exam attempt
        full_name(str): name to be used as verified name
        profile_name(str): user's current profile name
        is_practice_exam(boolean): if the exam attempt is for a practice exam
        is_proctored(boolean): if the exam attempt is for a proctored exam
        backend_supports_onboarding(boolean): if the exam attempt is for an exam with a backend that supports onboarding
    """

    # We only care about updates from onboarding exams, or from non-practice proctored exams with a backend that
    # does not support onboarding. This is because those two event types are guaranteed to contain verification events,
    # whereas timed exams and proctored exams with a backend that does support onboarding are not guaranteed
    is_onboarding_exam = is_practice_exam and is_proctored and backend_supports_onboarding
    reviewable_proctored_exam = is_proctored and not is_practice_exam and not backend_supports_onboarding
    if not (is_onboarding_exam or reviewable_proctored_exam):
        return

    trigger_status = VerifiedNameStatus.trigger_state_change_from_proctoring(status)

    # only trigger celery task if status is relevant to name affirmation
    if trigger_status:
        proctoring_update_verified_name_task.delay(
            attempt_id,
            user_id,
            trigger_status,
            full_name,
            profile_name,
        )
    else:
        log.info('VerifiedName: proctoring_attempt_handler will not trigger Celery task for user %(user_id)s '
                 'with profile_name %(profile_name)s because of status %(status)s',
                 {
                     'user_id': user_id,
                     'profile_name': profile_name,
                     'status': status,
                 }
                 )


def proctoring_delete_handler(sender, instance, signal, **kwargs):  # pylint: disable=unused-argument
    """
    Receiver for proctoring attempt deletions

    Args:
        attempt_id(int): ID associated with the proctoring attempt
    """
    proctoring_attempt_id = instance.id
    log.info(
        'VerifiedName: proctoring_delete_handler triggering Celery task for '
        'proctoring_attempt_id=%(proctoring_attempt_id)s',
        {
            'proctoring_attempt_id': proctoring_attempt_id,
        }
    )
    delete_verified_name_task.delay(None, proctoring_attempt_id)



"""
Python API for edx_name_affirmation.
"""

import logging

from django.core.exceptions import ObjectDoesNotExist

from edx_name_affirmation.exceptions import (
    VerifiedNameAttemptIdNotGiven,
    VerifiedNameDoesNotExist,
    VerifiedNameEmptyString,
    VerifiedNameMultipleAttemptIds
)
from edx_name_affirmation.models import VerifiedName, VerifiedNameConfig
from edx_name_affirmation.statuses import VerifiedNameStatus

log = logging.getLogger(__name__)


def create_verified_name(
    user, verified_name, profile_name, verification_attempt_id=None,
    proctored_exam_attempt_id=None, platform_verification_attempt_id=None,
    status=VerifiedNameStatus.PENDING,
):
    """
    Create a new `VerifiedName` for the given user.

    Arguments:
        * `user` (User object)
        * `verified_name` (str): Representative of the name on the user's physical ID card.
        * `profile_name` (str): A snapshot of either 1) the user's pending name change if given
          or 2) the existing name on the user's profile.
        * `verification_attempt_id` (int): Optional reference to an external ID verification
          attempt.
        * `proctored_exam_attempt_id` (int): Optional reference to an external proctored exam
          attempt.
        * `platform_verification_attempt_id` (int): Optional reference to a platform defined
          verification attempt.
        * `is_verified` (bool): Optional, defaults False. This should determine whether the
          verified_name is valid for use with ID verification, exams, etc.
    """
    # Do not allow empty strings
    if verified_name == '':
        raise VerifiedNameEmptyString('verified_name', user.id)
    if profile_name == '':
        raise VerifiedNameEmptyString('profile_name', user.id)

    # Only link to one attempt
    if sum(map(bool, [proctored_exam_attempt_id, verification_attempt_id, platform_verification_attempt_id])) > 1:
        err_msg = (
            'Attempted to create VerifiedName for user_id={user_id}, but at least two different '
            'external attempt IDs were given. Only one may be used. '
            'verification_attempt_id={verification_attempt_id}, '
            'proctored_exam_attempt_id={proctored_exam_attempt_id}, '
            'platform_verification_attempt_id={platform_verification_attempt_id}, '
            'status={status}'.format(
                user_id=user.id, verification_attempt_id=verification_attempt_id,
                proctored_exam_attempt_id=proctored_exam_attempt_id, status=status,
                platform_verification_attempt_id=platform_verification_attempt_id,
            )
        )
        raise VerifiedNameMultipleAttemptIds(err_msg)

    VerifiedName.objects.create(
        user=user,
        verified_name=verified_name,
        profile_name=profile_name,
        verification_attempt_id=verification_attempt_id,
        proctored_exam_attempt_id=proctored_exam_attempt_id,
        platform_verification_attempt_id=platform_verification_attempt_id,
        status=status,
    )

    log_msg = (
        'VerifiedName created for user_id={user_id}. '
        'verification_attempt_id={verification_attempt_id}, '
        'proctored_exam_attempt_id={proctored_exam_attempt_id}, '
        'status={status}'.format(
            user_id=user.id, verification_attempt_id=verification_attempt_id,
            proctored_exam_attempt_id=proctored_exam_attempt_id, status=status,
        )
    )
    log.info(log_msg)


def get_verified_name(user, is_verified=False, statuses_to_exclude=None):
    """
    Get the most recent VerifiedName for a given user.

    Arguments:
        * `user` (User object)
        * `is_verified` (bool): Optional, set to True to ignore entries that are not
          verified.
        * `statuses_to_exclude` (list): Optional list of statuses to filter out. Only
          relevant if `is_verified` is False.

    Returns a VerifiedName object.
    """
    verified_name_qs = VerifiedName.objects.filter(user=user).order_by('-created')

    if is_verified:
        return verified_name_qs.filter(status=VerifiedNameStatus.APPROVED.value).first()

    if statuses_to_exclude:
        return verified_name_qs.exclude(status__in=statuses_to_exclude).first()

    return verified_name_qs.first()


def delete_verified_name(verified_name_id):
    """
    Delete a VerifiedName.
    """
    try:
        verified_name = VerifiedName.objects.get(id=verified_name_id)
        verified_name.delete()
    except ObjectDoesNotExist as exc:
        err_msg = (
            'Attempted to delete verified_name_id={verified_name_id}'
            'but it does not exist'.format(verified_name_id=verified_name_id)
        )
        raise VerifiedNameDoesNotExist(err_msg) from exc


def get_verified_name_history(user):
    """
    Return a QuerySet of all VerifiedNames for a given user, ordered by the date created from
    most recent.

    Arguments:
        * `user` (User object)
    """
    return VerifiedName.objects.filter(user=user).order_by('-created')


def update_verified_name_status(
    user, status, verification_attempt_id=None, proctored_exam_attempt_id=None, platform_verification_attempt_id=None,
):
    """
    Update the status of a VerifiedName using the linked ID verification, exam attempt ID, or platform defined
    verification attempt ID. Only one of these should be specified.

    Arguments:
        * user (User object)
        * status (Verified Name Status)
        * verification_attempt_id (int): Optional reference to an external ID verification
          attempt.
        * proctored_exam_attempt_id (int): Optional reference to an external proctored exam
          attempt.
        * platform_verification_attempt_id (int): Optional reference to a platform defined
          verification attempt.
    """
    filters = {'user': user}

    if sum(map(bool, [proctored_exam_attempt_id, verification_attempt_id, platform_verification_attempt_id])) > 1:
        err_msg = (
            'Attempted to update the status for a VerifiedName, but at least two different '
            'attempt IDs were given. verification_attempt_id={verification_attempt_id}, '
            'proctored_exam_attempt_id={proctored_exam_attempt_id},'
            'platform_verification_attempt_id={platform_verification_attempt_id}'.format(
                verification_attempt_id=verification_attempt_id,
                proctored_exam_attempt_id=proctored_exam_attempt_id,
                platform_verification_attempt_id=platform_verification_attempt_id,
            )
        )
        raise VerifiedNameMultipleAttemptIds(err_msg)

    if verification_attempt_id:
        filters['verification_attempt_id'] = verification_attempt_id
    elif proctored_exam_attempt_id:
        filters['proctored_exam_attempt_id'] = proctored_exam_attempt_id
    elif platform_verification_attempt_id:
        filters['platform_verification_attempt_id'] = platform_verification_attempt_id
    else:
        err_msg = (
            'Attempted to update the status for a VerifiedName, but no '
            'verification_attempt_id or proctored_exam_attempt_id was given.'
        )
        raise VerifiedNameAttemptIdNotGiven(err_msg)

    verified_name_obj = VerifiedName.objects.filter(**filters).order_by('-created').first()

    if not verified_name_obj:
        err_msg = (
            'Attempted to update status={status} for a VerifiedName, but one does '
            'not exist for the given attempt ID. verification_attempt_id={verification_attempt_id}, '
            'proctored_exam_attempt_id={proctored_exam_attempt_id}'.format(
                status=status,
                verification_attempt_id=verification_attempt_id,
                proctored_exam_attempt_id=proctored_exam_attempt_id,
            )
        )
        raise VerifiedNameDoesNotExist(err_msg)

    verified_name_obj.status = status
    verified_name_obj.save()

    log_msg = (
        'Updated status={status} for VerifiedName belonging to user_id={user_id}. '
        'verification_attempt_id={verification_attempt_id}, '
        'proctored_exam_attempt_id={proctored_exam_attempt_id}'.format(
            status=status,
            user_id=verified_name_obj.user.id,
            verification_attempt_id=verification_attempt_id,
            proctored_exam_attempt_id=proctored_exam_attempt_id,
        )
    )
    log.info(log_msg)

    return verified_name_obj


def create_verified_name_config(user, use_verified_name_for_certs=None):
    """
    Create verified name configuration for the given user.
    Arguments:
        * `user` (User object)
        * `use_verified_name_for_certs` (bool): If True, certificates will prioritize
          the user's verified name over their profile name.
    """
    fields = {'user': user}

    # Default to the values from the most recent config
    existing_config = VerifiedNameConfig.objects.filter(user=user).order_by('-change_date').first()
    if existing_config:
        fields['use_verified_name_for_certs'] = existing_config.use_verified_name_for_certs

    if use_verified_name_for_certs is not None:
        fields['use_verified_name_for_certs'] = use_verified_name_for_certs

    config_obj = VerifiedNameConfig.objects.create(**fields)

    log_msg = (
        'Created verified name config for user_id={user_id} with '
        'use_verified_name_for_certs={use_verified_name_for_certs}'.format(
            user_id=config_obj.user.id,
            use_verified_name_for_certs=config_obj.use_verified_name_for_certs,
        )
    )
    log.info(log_msg)


def should_use_verified_name_for_certs(user):
    """
    Returns a boolean describing whether the user has opted to use their verified
    name over their profile name for certificates.
    Arguments:
        * `user` (User object)
    """
    config_obj = VerifiedNameConfig.current(user)
    return config_obj.use_verified_name_for_certs
